@@ -9,6 +9,7 @@ set -u
 set -o pipefail
 
 identify_library_abi=
+libcurl_compat_setup=
 
 # Check if set, which is normally done by steam.sh, but will not be set when invoked directly
 if [ -z "${STEAM_ZENITY+x}" ]; then
@@ -37,6 +38,22 @@ debug ()
     fi
 }
 
+progress ()
+{
+    local percent="$1"
+    local zenity_progress="$2"
+
+    # Suppress "broken pipe" error message if any, and fall back to
+    # TUI progress reporting on stderr.
+    if [ "$zenity_progress" = true ] && echo "$percent" 2>/dev/null; then
+        return 0
+    elif [ "$percent" = 100 ]; then
+        echo " 100%   " >&2 || :
+    else
+        printf '\r %d%%    \r' "$percent" >&2 || :
+    fi
+}
+
 pin_newer_runtime_libs ()
 {
     # Set separator to newline just for this function
@@ -56,6 +73,12 @@ pin_newer_runtime_libs ()
         zenity_progress=$2
     fi
 
+    if [ "$zenity_progress" = true ]; then
+        # Don't get killed by SIGPIPE if zenity has crashed or otherwise
+        # failed
+        trap '' PIPE || :
+    fi
+
     # Associative array; indices are the SONAME, values are final path
     local -A host_libraries_32
     local -A host_libraries_64
@@ -85,6 +108,8 @@ pin_newer_runtime_libs ()
     local soname_fullpath
     local soname_symlink
 
+    rm -rf "$steam_runtime_path/libcurl_compat_32"
+    rm -rf "$steam_runtime_path/libcurl_compat_64"
     rm -rf "$steam_runtime_path/pinned_libs_32"
     rm -rf "$steam_runtime_path/pinned_libs_64"
 
@@ -93,11 +118,7 @@ pin_newer_runtime_libs ()
     # The ldconfig, 60% of the total
     # The pinning, 35% of the total
     # And at the end we just print 100%
-    if [ "$zenity_progress" = true ]; then
-        echo 4
-    else
-        printf '\r 4%%    \r' >&2 || :
-    fi
+    progress 4 "$zenity_progress"
 
     if [[ -n "$identify_library_abi" ]]; then
         local executable_output_array
@@ -146,11 +167,7 @@ pin_newer_runtime_libs ()
 
         for ldconfig_output in "${ldconfig_output_array[@]}"
         do
-            if [ "$zenity_progress" = true ]; then
-                echo $(( ( 60 * n_done ) / ldconfig_num + 4 ))
-            else
-                printf '\r %d%%    \r' $(( ( 60 * n_done ) / ldconfig_num + 4 )) >&2 || :
-            fi
+            progress $(( ( 60 * n_done ) / ldconfig_num + 4 )) "$zenity_progress"
             (( n_done=n_done+1 ))
             # If line starts with a leading / and contains :, it's a new path prefix
             if [[ "$ldconfig_output" =~ ^/.*: ]]
@@ -201,15 +218,21 @@ pin_newer_runtime_libs ()
         done
     fi
 
-    if [ "$zenity_progress" = true ]; then
-        echo 64
-    else
-        printf '\r 64%%    \r' >&2 || :
-    fi
+    progress 64 "$zenity_progress"
 
     mkdir "$steam_runtime_path/pinned_libs_32"
     mkdir "$steam_runtime_path/pinned_libs_64"
 
+    # Give steamrt a chance to fix libcurl ABI conflicts in a cleverer way.
+    # This will only work for glibc >= 2.30, but if it does work, it will
+    # create a libcurl_compat_${bitness}/libcurl.so.4 that is better than
+    # anything we can do from this shell script. This relies on tricky
+    # implementation details of glibc, so we're not using it by default yet
+    # (see run.sh for the opt-in mechanism).
+    if [[ -x "$libcurl_compat_setup" ]] && "$libcurl_compat_setup" --runtime-optional "$steam_runtime_path"; then
+        debug "run.sh can optionally use shim library to support more than one libcurl ABI"
+    fi
+
     if [[ -n "$identify_library_abi" ]]; then
         mapfile -t output_array < <($identify_library_abi --directory "$steam_runtime_path" --skip-unversioned  2>/dev/null)
     else
@@ -234,11 +257,7 @@ pin_newer_runtime_libs ()
                 ;;
         esac
 
-        if [ "$zenity_progress" = true ]; then
-            echo $(( ( 35 * n_done ) / find_num + 64 ))
-        else
-            printf '\r %d%%    \r' $(( ( 35 * n_done ) / find_num + 64 )) >&2 || :
-        fi
+        progress $(( ( 35 * n_done ) / find_num + 64 )) "$zenity_progress"
         (( n_done=n_done+1 ))
 
         if [[ -n "$identify_library_abi" ]]; then
@@ -274,76 +293,58 @@ pin_newer_runtime_libs ()
 
         soname=${soname_symlink##*/}
 
-        # If we had entries in our arrays, get them
-        if [[ -n "${host_libraries_32[$soname]+isset}" ||
-              -n "${host_libraries_64[$soname]+isset}" ]]
+        if [[ -z $soname_details ]]; then
+            soname_details=$(file -L "$final_library")
+        fi
+
+        if [[ $soname_details == *"32-bit"* || $soname_details == "i386-linux-gnu" ]]
         then
-            if [[ -z $soname_details ]]; then
-                soname_details=$(file -L "$final_library")
-            fi
-
-            if [[ $soname_details == *"32-bit"* || $soname_details == "i386-linux-gnu" ]]
+            if [ -n "${host_libraries_32[$soname]+isset}" ]
             then
-                if [ -n "${host_libraries_32[$soname]+isset}" ]
-                then
-                    host_soname_symlink=${host_libraries_32[$soname]}
-                fi
-                bitness="32"
-            elif [[ $soname_details == *"64-bit"* || $soname_details == "x86_64-linux-gnu" ]]
-            then
-                if [ -n "${host_libraries_64[$soname]+isset}" ]
-                then
-                    host_soname_symlink=${host_libraries_64[$soname]}
-                fi
-                bitness="64"
+                host_soname_symlink=${host_libraries_32[$soname]}
             fi
-       fi
+            bitness="32"
+        elif [[ $soname_details == *"64-bit"* || $soname_details == "x86_64-linux-gnu" ]]
+        then
+            if [ -n "${host_libraries_64[$soname]+isset}" ]
+            then
+                host_soname_symlink=${host_libraries_64[$soname]}
+            fi
+            bitness="64"
+        fi
 
-        # Do we have a host library found for the same SONAME?
-        if [[ ! -f $host_soname_symlink || $bitness == "unknown" ]]; then continue; fi
-
-        host_library=$(readlink -f "$host_soname_symlink")
-
-        if [[ ! -f $host_library ]]; then continue; fi
-
-        #log $soname ${host_libraries[$soname]} $r_lib_major $r_lib_minor $r_lib_third
-
-        # Pretty sure the host library already matches, but we need the rematch anyway
-        if [[ ! $host_library =~ .*\.so.([[:digit:]]+).([[:digit:]]+).([[:digit:]]+)$ ]]; then continue; fi
-
-        h_lib_major=$((10#${BASH_REMATCH[1]}))
-        h_lib_minor=$((10#${BASH_REMATCH[2]}))
-        h_lib_third=$((10#${BASH_REMATCH[3]}))
+        if [[ $bitness == "unknown" ]]; then continue; fi
 
         runtime_version_newer="no"
 
-        if [[ $h_lib_major -lt $r_lib_major ]]; then
-            runtime_version_newer="yes"
-        fi
-
-        if [[ $h_lib_major -eq $r_lib_major && $h_lib_minor -lt $r_lib_minor ]]; then
-            runtime_version_newer="yes"
-        fi
-
-        if [[ $h_lib_major -eq $r_lib_major && $h_lib_minor -eq $r_lib_minor && $h_lib_third -lt $r_lib_third ]]; then
-            runtime_version_newer="yes"
-        fi
-
-        # There's a set of libraries that have to work together to yield a working dock
-        # We're reasonably convinced our set works well, and only pinning a handful would
-        # induce a mismatch and break the dock, so always pin all of these for Steam (32-bit)
-        if [[ $bitness == "32" ]]
-        then
-            if [[     "$soname" == "libgtk-x11-2.0.so.0"  || \
-                    "$soname" == "libdbusmenu-gtk.so.4"  || \
-                    "$soname" == "libdbusmenu-glib.so.4" || \
-                    "$soname" == "libdbus-1.so.3" ]]
-            then
-                runtime_version_newer="forced"
-            fi
-        fi
-
         case "$soname" in
+            (libdbusmenu-glib.so.4 | libdbusmenu-gtk.so.4)
+                # These two libraries are built from the same source package
+                # and assume that the other library is a matching version.
+                # If the host system has a newer 32-bit libdbusmenu-glib.so.4
+                # but does not have 32-bit libdbusmenu-gtk.so.4, then our
+                # usual pinning logic would result in using the new
+                # ldm-glib with the old ldm-gtk, a version mismatch with
+                # the symptom of breaking the menu that should appear when
+                # you right-click the Steam tray icon. Avoid this mismatch
+                # by always using the Steam Runtime's copy for the 32-bit
+                # Steam executable.
+                # https://github.com/ValveSoftware/steam-for-linux/issues/4795
+                if [ "$bitness" = 32 ]; then
+                    runtime_version_newer="forced"
+                fi
+                ;;
+
+            (libgtk-x11-2.0.so.0)
+                # This seems to be tied up with #4795 as well.
+                # Potentially related:
+                # https://github.com/ValveSoftware/steam-for-linux/issues/8577
+                # https://github.com/ValveSoftware/steam-for-linux/issues/9324
+                if [ "$bitness" = 32 ]; then
+                    runtime_version_newer="forced"
+                fi
+                ;;
+
             (libcurl.so.4)
                 # libcurl in the Steam Runtime is internally identified
                 # as libcurl.so.4, but with a symlink at libcurl.so.3
@@ -363,19 +364,93 @@ pin_newer_runtime_libs ()
                 # built against the Steam Runtime will expect the old SONAME
                 # and versioned symbols; make sure we use the Steam Runtime
                 # version.
+                #
+                # (If steamrt-libcurl-compat already fixed this for us, then
+                # it will have created the libcurl.so.4 symlink and as a
+                # result we'll never get here.)
+                runtime_version_newer="forced"
+                ;;
+
+            (libcurl-gnutls.so.4)
+                # Similar to the above, the GNUTLS variant of libcurl has
+                # Debian-specific symbol-versioning. We didn't have a
+                # problem with this until recently, because most OSs either
+                # don't ship a GNUTLS variant at all (Red Hat), ship it
+                # without versioned symbols (Arch), or are Debian-derived
+                # (Debian/Ubuntu), but apparently OpenMandriva has a
+                # non-Debian-compatible GNUTLS variant.
                 runtime_version_newer="forced"
                 ;;
         esac
 
+        # If the library is one of the ones we force, we need to
+        # do that even if we are not aware of an equivalent on the host
+        # system, because continuing uncertainty about the ABIs of these
+        # libraries means that it might have an unexpected SONAME.
+        # For instance, some versions of OpenMandriva's lib64curl-gnutls4
+        # contained a libcurl-gnutls.so.4 with SONAME libcurl.so.4.
+        if [ "$runtime_version_newer" = forced ]; then
+            log "Forced use of runtime version for $bitness-bit $soname"
+            ln -fns "$final_library" "$steam_runtime_path/pinned_libs_$bitness/$soname"
+            # For similar historical reasons, our libcurl*.so.3 are equivalent
+            # to libcurl*.so.4
+            case "$soname" in
+                (libcurl*.so.4)
+                    ln -fns "$final_library" "$steam_runtime_path/pinned_libs_$bitness/${soname%.4}.3"
+                    ;;
+            esac
+            continue
+        fi
+
+        # Do we have a host library found for the same SONAME?
+        if [[ ! -f $host_soname_symlink ]]; then continue; fi
+
+        host_library=$(readlink -f "$host_soname_symlink")
+
+        if [[ ! -f $host_library ]]; then continue; fi
+
+        #log $soname ${host_libraries[$soname]} $r_lib_major $r_lib_minor $r_lib_third
+
+        if [[ "$soname" = "libcrypt.so.1" \
+              && $host_library =~ libcrypt-2\.[[:digit:]]+\.so \
+              && $final_library =~ libcrypt\.so\.1\.* ]]
+        then
+            # Steam Runtime libxcrypt libcrypt.so.1 counts as newer than
+            # host glibc libcrypt.so.1, so behave as though the host version
+            # was arbitrarily old
+            debug "Treating libxcrypt libcrypt.so.1 as newer than host glibc version"
+            h_lib_major=0
+            h_lib_minor=0
+            h_lib_third=0
+        elif [[ $host_library =~ .*\.so.([[:digit:]]+).([[:digit:]]+).([[:digit:]]+)$ ]]; then
+            h_lib_major=$((10#${BASH_REMATCH[1]}))
+            h_lib_minor=$((10#${BASH_REMATCH[2]}))
+            h_lib_third=$((10#${BASH_REMATCH[3]}))
+        else
+            continue
+        fi
+
+        if [[ $h_lib_major -lt $r_lib_major ]]; then
+            runtime_version_newer="yes"
+        fi
+
+        if [[ $h_lib_major -eq $r_lib_major && $h_lib_minor -lt $r_lib_minor ]]; then
+            runtime_version_newer="yes"
+        fi
+
+        if [[ $h_lib_major -eq $r_lib_major && $h_lib_minor -eq $r_lib_minor && $h_lib_third -lt $r_lib_third ]]; then
+            runtime_version_newer="yes"
+        fi
+
+        if [[ "$soname" == libudev.so.0 && "$h_lib_major.$h_lib_minor.$h_lib_third" == 0.0.9999 ]]; then
+            # Work around https://github.com/archlinux/libudev0-shim/issues/4
+            debug "Treating libudev0-shim as newer than runtime version"
+            runtime_version_newer="no"
+        fi
+
         # Print to stderr because zenity is consuming stdout
         if [[ $runtime_version_newer == "yes" ]]; then
             log "Found newer runtime version for $bitness-bit $soname. Host: $h_lib_major.$h_lib_minor.$h_lib_third Runtime: $r_lib_major.$r_lib_minor.$r_lib_third"
-        elif [[ $runtime_version_newer == "forced" ]]; then
-            log "Forced use of runtime version for $bitness-bit $soname. Host: $h_lib_major.$h_lib_minor.$h_lib_third Runtime: $r_lib_major.$r_lib_minor.$r_lib_third"
-        fi
-
-        if [[ $runtime_version_newer == "yes" \
-              || $runtime_version_newer == "forced" ]]; then
             ln -fns "$final_library" "$steam_runtime_path/pinned_libs_$bitness/$soname"
             # Keep track of the exact version name we saw on the system at pinning time to check later
             >&2 echo "$host_soname_symlink" > "$steam_runtime_path/pinned_libs_$bitness/system_$soname"
@@ -385,25 +460,10 @@ pin_newer_runtime_libs ()
     done
 
     for bitness in 32 64; do
-        if [ -L "$steam_runtime_path/pinned_libs_$bitness/libcurl.so.4" ]; then
-            # The version of libcurl.so.4 in the Steam Runtime is actually
-            # binary-compatible with the older libcurl.so.3 in Debian/Ubuntu,
-            # so pin it under both names.
-            ln -fns libcurl.so.4 "$steam_runtime_path/pinned_libs_$bitness/libcurl.so.3"
-        fi
-
-        if [ -L "$steam_runtime_path/pinned_libs_$bitness/libcurl-gnutls.so.4" ]; then
-            # Similarly, libcurl-gnutls.so.4 is actually binary-compatible
-            # with libcurl-gnutls.so.3, so pin it under both names.
-            ln -fns libcurl-gnutls.so.4 "$steam_runtime_path/pinned_libs_$bitness/libcurl-gnutls.so.3"
-        fi
+        touch "$steam_runtime_path/pinned_libs_$bitness/done"
     done
 
-    if [ "$zenity_progress" = true ]; then
-        echo 100
-    else
-        echo " 100%   " >&2 || :
-    fi
+    progress 100 "$zenity_progress"
 }
 
 check_pins ()
@@ -425,8 +485,9 @@ check_pins ()
 
     pins_need_redoing="no"
 
-    # If we had the runtime previously unpacked but never ran the pin code, do it now
-    if [[ ! -d "$steam_runtime_path/pinned_libs_32" ]]
+    # If we had the runtime previously unpacked but never ran the pin code,
+    # or if a previous attempt failed or was cancelled, do it now
+    if ! [[ -e "$steam_runtime_path/pinned_libs_32/done" && -e "$steam_runtime_path/pinned_libs_64/done" ]]
     then
         pins_need_redoing="yes"
     fi
@@ -577,6 +638,7 @@ main ()
     esac
 
     identify_library_abi="$top/$arch/usr/bin/steam-runtime-identify-library-abi"
+    libcurl_compat_setup="$top/$arch/usr/bin/steam-runtime-libcurl-compat-setup"
 
     if [[ ! -x "$identify_library_abi" ]]; then
         identify_library_abi="$top/$arch/bin/steam-runtime-identify-library-abi"
