@@ -7,7 +7,6 @@
 
 import argparse
 import glob
-import io
 import logging
 import os
 import re
@@ -656,7 +655,6 @@ def main():
     logging.basicConfig()
     logging.getLogger().setLevel(logging.INFO)
 
-    config = {}
     os_release = OsRelease()
     missing_packages = []    # type: typing.List[str]
 
@@ -782,7 +780,8 @@ def main():
             cp = run_subprocess(argv, universal_newlines=True)
             return cp.returncode
         elif not is_root():
-            # We don't have root privileges, call again this script with pkexec/sudo
+            # We don't have root privileges, call again this script with
+            # pkexec/sudo
             print('The packages cache seems to be out of date')
             input('\nPress return to update the list of available packages: ')
 
